'*************************************
'*   zicog_demo v0.6                 *
'*   An emulation of the Zilog Z80   *
'*   (C) 2009 Michael Rychlik        *
'*************************************
'
'v0.0 -  1 Mar 2009 - Original version
'
'v0.1 -  4 Mar 2009 -
'
'v0.2 -  7 Mar 2009 -
'
'v0.3 -  7 Mar 2009 -
'
'v0.4 - 16 Mar 2009 - Replaced T8080 diagnostic with ZICOGTST.COM. This is a version of EXZ80DOC.COM from
'                     the SIMH AltairZ80 simulator with undocumented instructions removed and tests rearranged
'                     in execution time order (almost). T8080.COM no longer passes tests because due to the
'                     implementation of Z80 flag settings rather than 8080.
'                     Fixed up SIMH I/O simulation for console, disks on SD card etc.
'
'v0.5 - 18 Mar 2009 - 
'
'v0.6 - 30 Apr 2009 - Removed "outa" capability in output emulation.
'
'This file is a minimal example of how to set up and run the zicog z80 emulator
'
'' RR20090404   Fix required sections for compiling by homespun024.exe
''              sdspiqasm, PC_Keyboard, PC_Text, FrequencyCount all reguired org -> org 0
''              zicog: jmp requires # for line 374, and a description for jmp without # to say jmp/ret(indirect) 
''
'' +-------------------------------------------------------------------------------------------------------+
'' | Modifications to ZiCog v0.5 for use on Cluso's TriBladeProp PCB (Blade #2)                       v048 |
'' +-------------------------------------------------------------------------------------------------------+
'' |  Author:            "Cluso99" (Ray Rodrick)                                                           |
'' |  Copyright (c) 2009 "Cluso99" (Ray Rodrick)                                                           |
'' |  License            Modifications only under MIT License                                              |
'' +-------------------------------------------------------------------------------------------------------+
'' RR20090317   Cluso99: Modifications to ZiCog v0.5 for use on the TriBladeProp PCB (Blade #2)
'' RR20090317   start to use TriBladeProp & SRAM                             
''              Ram is only $4F00 - shouldn't it be $5000 ??                 
''              Remove heartbeat and FrequencyCounter.spin and fcounter.xxx calls
''              Remove sdspiqasm.spin and disk.xxx calls                     
''              Remove output LEDs                                           
'' RR20090318   add sram_start (ram offset)
''              Rom is $0180 long beginning $FF00 - shouldn't it be $0100 long ??
'' RR20090318   move from v0.4 to v0.5
'' RR20090318   _rr004 version
'' RR20090319   make 64KB CPM version
''              change msize 20 -> 64; disable free_ reservation in hub
'' RR20090320   _rr006 version: add SRAM disk
''              _rr007 complete and enable SRAM disk
'' RR20090321   _rr008 use new CBIOSX_triblade.COM
''              preload SRAM disk from cpm2.dsk using cpm2SD_sram_rr008 (truncates)
''              enabled readblock & writeblock (no clear on start). DIR -> "no file"
'' RR20090321   _rr009 incorporate cpm2SD_sram_rr008 to preload SRAM with cpm disk
''              short of stack space - disable part of print_regs debug code
''              _rr010 DIR says "no file"
''              _rr011 use sdsram_rr011 for loading xxxx.COM files to Z80RAM (64KB)
''              _rr015 preload SRAM disk first using cpm2SD_debug4 & SRAM_driver_015
'' RR20090331   _rr021 use TriBladeProp_Blade2_201.spin driver to load Z80 Ram and CPM Ram Disk
''              _rr022 bugfix (sector 127->128) Use Cluso's VB6 to load Ram from PC
''              _rr023 works loading SRAM and verified (using VB) for Z80 & Ram Disk 448_512KB
''                     gives A> prompt but DIR still does not work
''              _rr024 LoadSRAM optional constant
''              _rr025 the Ram Disk files have 137 byte sectors (not 128 as first thought) so fix in spin for now
''                     Hooray it does DIR :-)
''                     Warning, likely to require more stack space before continuing !!!
'' RR20090401   _rr026 start to load from microSD the SRAM 1MB "DRIVE_1M.RAM" (uses first 128 bytes of 512 byte sectors)
''              _rr027 remove the cpm code included files (z80ram tpa, ccp, etc) to freeup space
'' RR20090402   _rr028 add SD_Trainer_rr008 "FullDuplexSerialPlus" & "fsrwFemto" for microSD load to SRAM
''              _rr034 writing issue - try new driver _v202 (not the problem)
'' RR20090403   _rr035 (proved sram walking code ok, so some interaction with sd card drivers?)
''              _rr036 Fixed bug: SD card not tristating D0 - fixed sdspiFemto_rr001.spin
''              _rr040 Loading 1MB SRAM from "DRIVE_1M.RAM" on microSD and getting A>
''              _rr041 Does a DIR listing (SRAM loaded, so loading commented out)
'' RR20090404   _rr043 Migrate to ZiCog v0.006 & homespun compiler
'' ---------------------------------------------------------------------------------------------------------
'' RR20090404   _rr044 Migrate to ZiCog v0.006 & homespun compiler
''              _rr045 question to load SRAM from microSD, Fix ram_size
''              _rr045 Write working on _rr042(later), added here & works
''                     Note does not work with zicog_rr044 but does with zicog005_rr021
'' RR20090405   _rr046 Remove clear screen for loading from PropCMD - working & posted
''              _rr047 remove special prop UTF-16 characters
''              _rr048 done
'' RR20090407   _rr049 modify to use direct sector address for disk files on microSD card
''              _rr050 loading the 64KB+960KB using direct SD sector addressing
'' RR20090408   _rr051 convert keys
'' ---------------------------------------------------------------------------------------------------------
'' SKIP _rr052 & _rr053 (for now)
'' RR20090416    _rr055 (from _rr051) clear sram 0-64KB and only load $FF00-$FFFF DSKBOOT.COM
''                      now CPM should boot from the sram disk - does not work, so place jmp to $FF00 at $0000
'' RR20090416    _rr056 working (removed jmp $FF00 at $0000)
''               _rr057 get 2nd disk running (on SD card)
'' RR20090417    _rr058 continue
''               _rr059 continue
''               _rr060 continue (new SD files DSK_CPM2.DSK & DSK_CPM3.DSK) older formats incorrect
''                      loads ok but will not take kbd input due to restarts???
''               _rr061 continue (add #define sddisk)
''               _rr062 conflict (add #define sduseinit) prevent readsd from rebooting cog each time - works
''               _rr063 conflict (#define sddisk) - no kbd input at A> - now works, but resetting tbp2 cog !!
''                      but not using the sd disk data yet
''               _rr064 use external CPM2 disk on sd card (#define sdCPM2)
''               _rr065 (#define RamDisk) Note to fix resetting tbp2 from resetting when accessing sdcard
''                      working
''               _rr066 tidy code (tbp2 driver requires "idle" command instead of start/stop)
''                      add drive1base and 2 disks   - works
'' RR20090420    _rr067 try to do syscpm2 - works
'' RR20090421    _rr068 add drives C...H


'' NOTE: Require tbp2 "idle" command instead of start/stop

CON
#define TriBladeProp                                    'Uncomment for TriBladeProp hardware
'#define Display_Sectors                                'Uncomment to display sectors when starting
'#define debugfirsttime                                 'Uncomment to display the first sector read
#define Show_Hex                                        'Uncomment to display in hex, else display char
'#define enabledbgcnt                                   'Uncomment to do dbgcnt debug info
#define convertkeys                                     'converts F1>CtlZ, F2>CtlS, F3>CtlC
#define diskboot                                        'boot using cleared ram and $FF00 dskboot.com
'#define RamDisk                                        'use Ram Disk, else SD Card for CPM disk


#ifdef TriBladeProp
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

' microSD pins on TriBladeProp Blade #2                  
  spiDO     = 9  
  spiClk    = 28 
  spiDI     = 8  
  spiCS     = 14 

  LoadSRAM    = false                                   ' true = loads the SRAM from VB, false skips the load (already done)  
  Disk_Offset = 64*1024                                 ' constant to add to sector no to skip 64KB RAM                       

#else
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000
#endif

OBJ

#ifdef TriBladeProp
' cpu      : "zicog_rr044"                              'Zilog Z80 emulator in one COG
  cpu      : "zicog005_rr021"                           'Zilog Z80 emulator in one COG
' cpu      : "zicog005_rr021ss"                         'Zilog Z80 emulator in one COG - singlestep
  kb       : "PC_Keyboard"                              'You can also use Keyboard object for PS2-Keyboard                          
  TV       : "PC_Text"                                  'You can also use TV_Text or VGA_Text objects
  tbp2     : "TriBladeProp_Blade2_201"                  'Blade2 hardware driver 
  debug    : "FullDuplexSerialPlus"                     'reqd to load SRAM from microSD, then closed & kb + TV started
  sd       : "fsrwFemto"                                'SD Software used in FemtoBASIC
#else
  cpu      : "zicog"                                    'Zilog Z80 emulator in one COG
  disk     : "sdspiqasm"
  kb       : "PC_Keyboard"                              'You can also use Keyboard object for PS2-Keyboard                          
  TV       : "PC_Text"                                  'You can also use TV_Text or VGA_Text objects
  fcounter : "FrequencyCount"                           'Frequency counter to get execution speed
#endif


CON

'Z80 Program Status Word bits
sign_bit   = %10000000
zero_bit   = %01000000
aux_bit    = %00010000
parity_bit = %00000100
carry_bit  = %00000001


'I/O control block commands
io_cmd_out   = $01 
io_cmd_in    = $02
io_cmd_break = $03


'The following taken from the SIMH Altair simulator cfgccp.lib
'WARNING !!! Change msize is altair_mem also if you change it here !!!

#ifdef TriBladeProp
  msize   = 64           ' adjust per installed system memory (K bytes)
#else
  msize   = 20           ' adjust per installed system memory (K bytes)
#endif
  bioslen = $d00         ' adjust as bios changes are made (if necessary)
  ccplen  = $800         ' cp/m constant
  bdoslen = $e00         ' cp/m constant

' cp/m image size (rounded up to next 1k boundary)
cpmlen = (ccplen + bdoslen + bioslen + $3ff) / $400

ccpph   =     (msize - cpmlen) * 1024         ' ccp start address
bdosph  =     ccpph + ccplen                  ' bdos start address
biosph  =     ccpph + ccplen + bdoslen        ' bios start address

tpa     =     $100                            ' CP/M's Transient Program Area

bdos    = bdosph + 6                          ' Entry point for BDOS

ram_size = (1024 * msize) - $100              ' Excludes ROM
rom_start = $FF00


'The following taken from SIMH Altair simulator CBIOSX.MAC
ndisks  = 8                     ' total number of Altair disks
tracks  = 254                   ' number of tracks for regular drives
track1  = tracks+1              ' indicator for unknown track position
asecsiz = 137                   ' sector size Altair
csecsiz = $0080                 ' sector size CP/M



'       Address         Mode    Function
'       -------         ----    --------
'       selout          Out     Selects and enables controller and drive
'       statin          In      Indicates status of drive and controller
'       dskcon          Out     Controls disk function
'       secpos          In      Indicates current sector position of disk
'       dskwrit         Out     Write data
'       dskread         In      Read data

selout = 8                      ' port to select and enable controller and drive (OUT)
'       +---+---+---+---+---+---+---+---+
'       | C | X | X | X |   Device      |
'       +---+---+---+---+---+---+---+---+
'
'       C       = If this bit is 1, the disk controller selected by 'device' is
'                 cleared. If the bit is zero, 'device' is selected as the
'                 device being controlled by subsequent I/O operations.
'       X       = not used
'       Device  = value zero thru 15, selects drive to be controlled.

statin = 8                      ' port indicating status of drive and controller (IN)
'       +---+---+---+---+---+---+---+---+
'       | R | Z | I | X | X | H | M | W |
'       +---+---+---+---+---+---+---+---+
'
'       W - When 0, write circuit ready to write another byte.
'       M - When 0, head movement is allowed
'       H - When 0, indicates head is loaded for read/write
'       X - not used (will be 0)
'       I - When 0, indicates interrupts enabled (not used this simulator)
'       Z - When 0, indicates head is on track 0
'       R - When 0, indicates that read circuit has new byte to read

dskcon = 9                      ' port to control disc function (OUT)
'       +---+---+---+---+---+---+---+---+
'       | W | C | D | E | U | H | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - When 1, steps head IN one track
'       O - When 1, steps head OUT one track
'       H - When 1, loads head to drive surface
'       U - When 1, unloads head
'       E - Enables interrupts (ignored by this simulator)
'       D - Disables interrupts (ignored by this simulator)
'       C - When 1 lowers head current (ignored by this simulator)
'       W - When 1, starts Write Enable sequence:
'           W bit on device 'statin' (see above) will go 1 and data will be read from
'           port 'dskread' until 137 bytes have been read by the controller from
'           that port. The W bit will go off then, and the sector data will be written
'           to disk. Before you do this, you must have stepped the track to the desired
'           number, and waited until the right sector number is presented on
'           device 'secpos', then set this bit.

secpos = 9                      ' port to indicate current sector position of disk (IN)
'       As the sectors pass by the read head, they are counted and the
'       number of the current one is available in this register.
'
'       +---+---+---+---+---+---+---+---+
'       | X | X |  Sector Number    | T |
'       +---+---+---+---+---+---+---+---+
'
'       X               = Not used
'       Sector number   = binary of the sector number currently under the head, 0-31.
'       T               = Sector True, is a 1 when the sector is positioned to read or write.
'                         Note: MPR according to the CBIOSX this is 0 when sector is positioned.

dskwrit = 10                    ' port to write data (OUT)
dskread = 10                    ' port to read data (IN)

'       All I/O is via programmed I/O. Each device has a status port
'       and a data port. A write to the status port can select
'       some options for the device although the simulator only
'       recognizes the reset command (0x03).
'       A read of the status port gets the port status:
'
'       +---+---+---+---+---+---+---+---+
'       | X | X | X | X | X | X | O | I |
'       +---+---+---+---+---+---+---+---+
'
'       I - A 1 in this bit position means a character has been received
'               on the data port and is ready to be read.
'       O - A 1 in this bit means the port is ready to receive a character
'               on the data port and transmit it out over the serial line.
'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat = 16                    ' sio port 1 status port
condata = 17                    ' sio port 1 data port
punstat = 18                    ' sio port 2 status port
pundata = 19                    ' sio port 2 data port

simhport = $FE
srport   = $FF                  ' Altair front panel switch register port

ledport  = $00

' masks for disk controller (statin)
mhm     = $02                   ' head movement mask
mtzero  = $40                   ' head on track zero mask
mall    = $ff                   ' everything ok mask

' commands for disk controller (dskcon)
cstepin = $01                   ' step in command
cstepot = $02                   ' step out command
cload   = $04                   ' load head to drive surface command
cuload  = $08                   ' unload head from drive surface command
cwrseq  = $80                   ' 'start write enable sequence' command

' masks for SIO controller (constat, punstat)
mout    = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset  = 3                     ' reset command

' constants for hard disk port
hdskReset      = 1                      ' command to reset controller
hdskRead       = 2                      ' read command
hdskWrite      = 3                      ' write command
hdskport       = $fd                    ' control port for simulated hard disk
firstSector    = 17                     ' first sector to load
firstTrack     = 0                      ' from this track
firstDiskAddr  = 256*firstTrack+firstSector
sectors        = (ccplen+bdoslen)/csecsiz

'constants for SIMH port
simhreset = 14
stoptimer = 22
setbankselect = 12
hasbankedmemory = 18



DAT

'The Z80 Registers.
'Do not change the order of these.
'Registers are defined in pairs as words so as to enable them to be
'accessed individually as BYTE or in pairs as WORD.
'The ordering makes it easier to find a register specified by a 3-bit
'register field within an op code or to swap between main and
'alternate register sets. 
register_file word
bc_reg        word      0
de_reg        word      0
hl_reg        word      0
af_reg        word      0
bc_reg_alt    word      0 
de_reg_alt    word      0
hl_reg_alt    word      0
af_reg_alt    word      0
im_reg        word      0
ix_reg        word      0
iy_reg        word      0
sp_reg        word      0
pc_reg        word      0
  
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0            
io_data       long      0


'The Z80 will start executing instructions from here
#ifdef TriBladeProp
ram_z80                                                    
'DOESN'T GET USED ANYMORE RR20090416
              'BYTE $C3, $00, $F2                          'Jump to CBIOSX (64K version RR20090319)
              BYTE $C3, $00, $FF                           'Jump to DSKBOOT and boot CP/M from ram disk
              BYTE 0[2]
              BYTE $C3, $06, $E4                           'Jump to BDOS entry (64K version RR20090319)
#else
ram_z80
              'OPTIONS: uncomment one (and only one) of the thre lines (BYTE ....) below as desired:  
              BYTE $C3, biosph & $FF, (biosph >> 8) & $FF 'Start CPM directly from RAM image, no disks (SD card) required for boot up
              'BYTE $C3, $00, $FF                          'Run the disk boot code from ROM space and boot CP/M from disk (SD card)
              'BYTE $C3, $00, $01                          'Jump to a program in the TPA
              
              'Place a jump to BDOS at location $0005
              BYTE 0[2]
              BYTE $C3, bdos & $FF, (bdos >> 8) & $FF

              'Pad out with zeros until the start of the Transient Program Area at $100
              BYTE 0[$0100 - 8]

tpa_          'The Transient Program Area (TPA)
              file "zicogtst.com"
              
              'Empty memory between a program in TPA and CPM
free_         BYTE 0[ccpph - tpa - (@free_ - @tpa_)]

              'Console Command Processor compiled from original Digital Research source.
ccp_          file "CCP.COM"

              'Basic Disk Operating System compiled from original Digital Resarch source
bdos_         file "BDOS.COM"

              'Customized BIOS of the SIMH Altair Emulator Project
cbiosx_       file "CBIOSX.COM"
ram_z80_end
              'The Altair boot rom. This read as rom located $FF00 by the cpu emulator
              'This file is the result of assembling DSKBOOT.MAC inside the SIMH Altair 8800 simulator.
rom_z80       file "DSKBOOT.COM"
#endif


VAR     

'Parameter list for CPU objects, do not change order.
long reg_base       'Address (in HUB)of Z80 register file
long ram_base       'Address (in HUB) of the Z80 RAM area
long ram_len        'Size of Z80 RAM area, assummed to start at zero in Z80 address space.
long rom_base       'Address (in HUB) of the Z80 ROM area
long rom_addr       'Address in Z80 RAM area of the start of ROM.
long io_base        'Address (in prop RAM) of the I/O contol interface
long cpu_control_ptr'Quick hack at single step control
long opcode_ptr     'Place where a fetched opcode lives.

word err            'Error code return value
long cpu_control

long frequency      'Frequency counter output goes here

long switch_reg 
long simhcmd

byte altair_drive        'The Altair disk drive currently selected, -1 = none
byte altair_track[16]    'The current track position on each Altair disk drive
byte altair_sector       'The current sector on an altair disk
byte altair_bidx         'Byte index into Altair disk sectors
byte altair_write_enable 'Bytes can be written to the Altair disk when = 1

byte disk_buff[512]
long sd_block

byte key

long dbgcnt
long firsttime
long sector

#ifdef TriBladeProp
'For SRAM CPM Disk Loading
  long  filename
  long  srcname
  byte  buffer[16]
  long  sectorsize
'For microSD routines
  long  ioControl[2]               ' SD parameters
  long  drive_base[16]             ' SD base address of drives 0..15 sector 0
  long  blockno                    ' blockno within file
  long  sdsector                   ' sector no 
  long  count                      ' count of chars to read
  byte  buff[8192]                 ' SD buffer
#endif


PUB start

  waitcnt(clkfreq*3 + cnt)      'delay (so you can get PropTerminal active)     <=== 3 secs

#ifdef TriBladeProp
'------------------------------------------------------------------------------
' Load SRAM with Z80 RAM (64KB) & CPM Ram Disk (960KB)
'------------------------------------------------------------------------------
  LoadSRAMfromSD                                        'load from microSD
#endif

'------------------------------------------------------------------------------
' ZiCog Emulation starts here
'------------------------------------------------------------------------------
#ifdef TriBladeProp
  reg_base        := @register_file
  ram_base        := $0000                              ' $0000
  ram_len         := ram_size
  rom_base        := $0000                              ' $0000
  rom_addr        := rom_start                  
  io_base         := @io_cmd_block
  cpu_control_ptr := @cpu_control
  switch_reg      := $08                     '$08 is the switch setting SIMH recommends for 4k Basic etc to run     
  firsttime       := 0                       'if 0 display first disk sector read by Z80
#else
  reg_base        := @register_file
  ram_base        := @ram_z80
  ram_len         := ram_size
  rom_base        := @rom_z80
  rom_addr        := rom_start
  io_base         := @io_cmd_block
  cpu_control_ptr := @cpu_control
  switch_reg      := $08                     '$08 is the switch setting SIMH recommends for 4k Basic etc to run     
  firsttime       := 0                       'if 0 display first disk sector read by Z80
#endif

  'Start the frequency counter
#ifdef TriBladeProp
#else
  fcounter.start(@frequency)
#endif

  'Start the terminal
  TV.start(12)
#ifdef TriBladeProp
  TV.str(string("ZiCog v0.006 on the TriBladeProp v0.06x",13))
#else
  TV.str(string("ZICOG v0.006",13))
#endif

  'Start the keyboard
  TV.str (string("Starting keyboard ... "))
  err := \kb.start(26, 27)
  DisplayError(not err)


#ifdef TriBladeProp
#else
  'Start the disk
  TV.str (string("Starting disks (A:B:C:D:E:F:G:H:) ... "))
  err := \disk.start(0)
  DisplayError(err)
#endif

  'Start the TriBladeProp driver
#ifdef TriBladeProp
  TV.str (string("Starting TriBladeProp driver ... "))
  err := tbp2.start
  DisplayError(err)
#endif

#ifdef Display_Sectors
'------------------------------------------------------------------------------
' Display any sector
'------------------------------------------------------------------------------
  TV.str(string("Press F1 to display next sector.. '???' to continue..", 13))
  sector := 0
  repeat                           
    repeat until kb.key == $D0
    DisplaySector(sector)
    sector++
#endif

#ifdef TriBladeProp
#else
  TV.str(string("RAM base = "))
  TV.hex(ram_base, 4)
  TV.out(13)

  TV.str(string("ROM base = "))
  TV.hex(rom_base, 4)
  TV.out(13)
#endif

'------------------------------------------------------------------------------
' Start Z80 emulation
'------------------------------------------------------------------------------
  TV.str(string("Starting Z80 emulation..."))
  err := \cpu.start(@reg_base)
  DisplayError(not err)

  'Start the Altair I/O simulation
  service_io

#ifdef TriBladeProp
'------------------------------------------------------------------------------
' TriBladeProp Routines
'------------------------------------------------------------------------------
PRI LoadSRAMfromSD | r, i, j, k
'------------------------------------------------------------------------------
' Load SRAM image from microSD: Z80 64KB + 960KB Ram Disk
'------------------------------------------------------------------------------

  debug.start(31,30,0,115200)   '' Open serial communication with PST           <=== 115200 (was 57600)

'  debug.tx(0)                                           'clear screen
  debug.str(string(13,13))                               'newline
  debug.str(string("TriBladeProp: Load SRAM from microSD (Y/N)? "))
  repeat
    k := debug.rx
    debug.tx(k)                                         'echo
    case k
      "y", "Y" : k := 1
      "n", "N" : k := 2
      other    : k := 0
  while k == 0

  blockno := 0
  count := 8192 
  debug.str (string(13,"Starting disks... "))
  FindSDblock                                           'find the address of the SD file block

  if k == 2                                             'check if we are loading sram?
    debug.tx(13)
    waitcnt(clkfreq/100 + cnt)
    debug.stop
    return 0

  debug.str(string(13,"Loading SRAM... "))
  repeat
    ReadSDblock                                         'read from SD
    WriteSRAMblock                                      'write to sram
    blockno++
    debug.tx(".")
  while blockno < 128                                   'eof? (all 1MB)

  debug.str(string(" Loaded.",13))
  waitcnt(clkfreq/4 + cnt)
  debug.stop                                            'stop the serial

PRI FindSDblock | i, n, r                               'read blockno of file for count into buffer

  r := sd.start(@ioControl)                             'start the SD routines
  check_error(not r)
  r := sd.mount(spiDO,spiClk,spiDI,spiCS)               'mount the SD
  check_error(r)

  r := sd.popen(string("DRIVE_1M.RAM"),"r")             'Open file for read (contains sram image)
  check_error(r)
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  check_error(r)
  sdsector := ioControl[0] - 1                          'get the sectorno of the file (returns the sector+1)

{
  r := sd.popen(string("DSK_CPM2.DSK"),"r")             'Open file for read (CPM 2.2 image)
' r := sd.popen(string("DRV_CPM2.DSK"),"r") invalid format             'Open file for read (CPM 2.2)
' r := sd.popen(string("DRIV_CTR.CPM"),"r")             'Open file for read (contains sector count)
' r := sd.popen(string("DRIVE_1M.RAM"),"r")             'Open file for read (contains sram image)
  check_error(r)
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  check_error(r)
  drive0base := ioControl[0] - 1                        'get the sectorno of the file (returns the sector+1)
                
  r := sd.popen(string("DRIVE__B.CPM"),"r")             'Open file for read (blank image)
' r := sd.popen(string("DSK_CPM3.DSK"),"r")             'Open file for read (CPM3 image)
  check_error(r)
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  check_error(r)
  drive1base := ioControl[0] - 1                        'get the sectorno of the file (returns the sector+1)
}

  FindDriveBase(0, string("DRIVE__A.DSK"))              'set the SD card drive base(s)
  FindDriveBase(1, string("DRIVE__B.DSK"))              'set the SD card drive base(s)
  FindDriveBase(2, string("DRIVE__C.DSK"))              'set the SD card drive base(s)
  FindDriveBase(3, string("DRIVE__D.DSK"))              'set the SD card drive base(s)
  FindDriveBase(4, string("DRIVE__E.DSK"))              'set the SD card drive base(s)
  FindDriveBase(5, string("DRIVE__F.DSK"))              'set the SD card drive base(s)
  FindDriveBase(6, string("DRVCPM_2.DSK"))              'set the SD card drive base(s)
  FindDriveBase(7, string("DRVCPM_3.DSK"))              'set the SD card drive base(s)
                
  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  check_error(r)

PRI FindDriveBase(n, st) | r                            'set the SD drive_base for the drive no
  debug.tx($41 + n)
  debug.tx(":")                                         'A:B:C:D:E:F:G:H
  r := sd.popen(st,"r")                                 'Open file for read (blank image)
  check_error(r)
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  check_error(r)
  drive_base[n] := ioControl[0] - 1                     'get the sectorno of the file (returns the sector+1)


PRI ReadSDblock | i, n, r                               'read blockno of file for count into buffer

  r := sd.initSDCard(spiDO,spiClk,spiDI,spiCS)          'init SD card, but do not restart cog
  check_error(r)

' this method relies on the SD file using contiguous sectors !!!
  r := sd.readSDCard(sdsector, @buff, count)            'get a block (ensure buff big enough!)
  check_error(r)
  sdsector := ioControl[0]                              'get the next sectorno

  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  check_error(r)


PRI WriteSRAMblock | i, r, mask, in

  r := tbp2.start
  check_error(r)

#ifdef diskboot
  'clears ram $0000-$EFFF, loads $FF00-$FFFF, loads sram disk above
  'remember we are reading 8KB blocks, so clear the seven blocks and 8KB-128B of eighth block
  if blockno < 8
    debug.dec(blockno)
    if blockno == 7
      repeat i from 0 to count-1 -(256)                 'clear below $FF00
        buff[i] := 0                                    'clear the buffer
    else
      repeat i from 0 to count-1
        buff[i] := 0                                    'clear the buffer
'     if blockno == 0
'       buff[0] := $C3                                  'force jmp to $FF00 (faster)
'       buff[1] := $00
'       buff[2] := $FF
#endif

  'write the block out to sram
  r := tbp2.DoCmd("W", @buff, blockno*count, count)    'write to sram
  check_error(r)                
'  show("w")                                  'show part of the block

  r := tbp2.stop
  check_error(r)


PRI preset_137
' Preset the extra 9 bytes (3 before, 6 after) the 128 byte sector
  disk_buff[0] := $A7                           ' add pre & post bytes to make 137 byte sector
  disk_buff[1] := $0E
  disk_buff[2] := $D2
  disk_buff[131] := $B7
  disk_buff[132] := $C8
  disk_buff[133] := $CD
  disk_buff[134] := $A5
  disk_buff[135] := $2B
  disk_buff[136] := $00

PRI show(s) | i, addrx
  addrx := blockno * count
  debug.hex(addrx,6)
  debug.tx(" ")
  repeat i from 0 to 15                                    'only display first xx chars per block
#ifdef Show_Hex
    debug.hex(buff[i],2)
    debug.tx(" ")
#else
    debug.tx(buff[i])
#endif
  debug.tx(13)

PRI Check_Error(r)
  if r == -1
    debug.str(string("***Error*** $"))
    debug.hex(r,2)
    sd.unmount
    sd.stop
    debug.str(string(13,"***HALTED***",13))
    repeat                      '<==============  LOOP HERE

'------------------------------------------------------------------------------
#endif


PRI CheckError(r)
  if r <> 0 '== -1
    TV.str(string("***Error*** $"))
    TV.hex(r,2)
    sd.stop
    tbp2.stop
    debug.str(string(13,"***HALTED***",13))
    repeat                      '<==============  LOOP HERE

PRI DisplayError(errx)
  if errx
    TV.str(string(" ***Failed*** $"))
    TV.hex(errx,2)
  else
    TV.str(string(" Passed "))
  TV.str(string(13))

'------------------------------------------------------------------------------
' Service I/O
'------------------------------------------------------------------------------
PRI service_io | now
'Altair I/O simulation service loop
  now := cnt 
  repeat
    case io_command
      io_cmd_break:                              'CPU break 
        on_break    
      io_cmd_in:                                 'An input operation
        on_input
      io_cmd_out:                                'An output operation
        on_output
{
    if cnt > now + 80_000_000
      TV.dec (frequency * 2)
      TV.out(13)
      now := cnt
}

PRI on_break
'Handle CPU break condition
'No, this never existed in an Altair (or SIMH simulator)
  print_regs
  'waitcnt (cnt + 80_000)
  repeat while kb.key == 0
  io_command := 0


PRI on_input
'Handle Z80 IO port reads
  case io_port
    srport:                                      'Altair front panel swithes register
      in_srport
    simhport:                                    'SIMH control port
      in_simhport          
    constat:                                     'Console status port.
      in_constat
    condata:                                     'Console data port
      in_condata
    statin:                                      'Altair disk status port
      in_statin
    secpos:                                      'Altair disk sector position port
      in_secpos
    dskread:                                     'Altair disk read port 
      in_dskread
    other:
      'TV.str(string("Read from unimplemented I/O Port $"))
      'TV.hex(io_port,2)
      'TV.out(13)
      'io_data := 0
  io_command := 0                                'Set flag ready for next io          

PRI in_srport
'Handle reads from the Altair front panel switch port
  io_data := switch_reg
  'TV.str(string("SR port read"))
  'TV.out(13)


PRI in_simhport
'Handle reads from the SIMH simulator status port
  if simhcmd == hasbankedmemory
    io_data :=  1
    simhcmd := 0

PRI in_constat
'Handle reads from the console status port
  if kb.gotkey == 0                              'Is there a key press ?
    io_data := %00000010                         'No: Return no char ready for input and output rdy.
  else
    key := kb.key                                'Yes: Get the key
    if key <> 0                                  'NULL character ?
'     TV.out("<")
'     TV.hex(key,4)
'     TV.out(">")
      case key                                   'No: Translate some keys for CP/M
        $00C8: key := $08                        'backspace
        $0263: key := $03                        'control-c
        $0265: key := $05                        'control-e
        $0268: key := $08                        'control-h
        $0270: key := $10                        'control-p
        $0272: key := $12                        'control-r
        $0273: key := $13                        'control-s
        $0275: key := $15                        'control-u
        $0278: key := $18                        'control-x
        $027A: key := $1A                        'control-z
        $00C9: key := $7F                        'rubout
        $00C2: key := $5E                        'up arrow
#ifdef convertkeys
        $D0  : key := $1A                        'F1 -> Cltr-Z
        $D1  : key := $13                        'F2 -> Ctrl-S
        $D2  : key := $03                        'F3 -> Ctrl-C
#endif  
          'if key => "a" and key =< "z"          'Use this for 4kBASIC which does not like lower case.
          '  key := key - $20
      io_data := %00000011                       'Return char ready for input and output rdy.
    else
      io_data := %00000010                       'Yes: Return no char ready for input and output rdy.
              
PRI in_condata
'Handle reads from the console data port
          io_data := key                         'Return the keyboard character

PRI in_statin
'Handle reads from the console status port
  io_data := 0
  'if altair_write_enable == 1                   'Ready to accept data ?
    'data |= %00000001
    'data |= %00000010                           'Head movement is always allowed !
    'data |= %00000100                           'Head is loaded for read/write
  if altair_track[altair_drive] <> 0             'Now on track zero ?
    io_data |= %01000000
    'data |= %10000000                           'New byte ready to read

PRI in_secpos
'Handle reads from the floppy disk sector position port
  altair_sector := altair_sector + 1
  if altair_sector > 31
    altair_sector := 0
  io_data := (altair_sector << 1) '| 1           'Return sector and positioned bit
                                                 'Note: MPR BIOS expects positioned bit = 0 when positioned
                                                 'despite what the BIOS comments say about it.
  altair_bidx := 0

PRI in_dskread | bix, r, iy
'Handle reads from the floppy disk read data port
  if altair_bidx == 0
    'Read a new block from SD card

#ifdef TriBladeProp
    preset_137

#ifdef RamDisk
    sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
    err := tbp2.DoCmd("R", @disk_buff[3], sd_block*128 + disk_offset, 128)     'works RR20090403
#else
'----------------------------------------------------------------------------------------------
'The following code resolves conflicts by stopping and restarting the drivers !!!!! needs fixing
'----------------------------------------------------------------------------------------------
    r := tbp2.stop                                      '<=== stop the tbp ram driver   !!!!
    CheckError(r)
     
    r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD
    CheckError(r)

    ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
{
    sd_block := (altair_track[altair_drive] * 32) + altair_sector
    if altair_drive == 0
      sd_block := sd_block + drive0base
    else
      sd_block := sd_block + drive1base
    err := sd.readSDCard(sd_block, @disk_buff[3], 128) 'read a sector
}
    sd_block := (drive_base[altair_drive]) + (altair_track[altair_drive] * 32) + altair_sector
    err := sd.readSDCard(sd_block, @disk_buff[3], 128) 'read a sector
'   CheckError(r)

    r := sd.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
    CheckError(r)
     
    r := tbp2.start                                       '<=== restart the tbp ram driver
    CheckError(r)
  '----------------------------------------------------------------------------------------------
#endif

#else
    sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
    err := \disk.readblock(sd_block, @disk_buff)
#endif

    if err
      TV.str(string("Disk Read "))
      DisplayError(err)
      DisplayDiskDTS

#ifdef enabledbgcnt
    if dbgcnt < 137
      DisplayDiskDTS
#endif

  io_data := disk_buff[altair_bidx]

#ifdef debugfirsttime
  if firsttime == 0
    firsttime := 1
    TV.str(string(13))
    TV.str(string("Sector="))
    TV.hex(sd_block,8)
    TV.str(string(" "))
    repeat bix from 0 to 136  '256
      TV.hex(disk_buff[bix],2)
      TV.str(string(" "))
    TV.str(string(13))
#endif

#ifdef enabledbgcnt
  if dbgcnt++ < 137
    TV.hex(io_data, 2)
    TV.out(" ")
#endif

  if altair_bidx < asecsiz
    altair_bidx++


PRI on_output
'Handle Z80 IO port writes
  case io_port
    ledport:                                     'An 8 bit parallel output with LEDs attached
      out_ledport          
    simhport:                                    'SIMH control port
      out_simhport
    condata:                                     'Console data port
      out_condata
    constat:                                     'Console status port
      out_constat
    punstat:
      out_punstat                                'Punch control port
    selout:                                      'Altair disk controller select port
      out_selout
    dskcon:                                      'Altair disk function port
      out_dskcon
    dskwrit:                                     'Altair disk write port
      out_dskwrit
    other:
      'TV.str(string("Write to unimplemented I/O Port $"))
      'TV.hex(io_port,2)
      'TV.out(13)
  io_command := 0                                 'Set flag ready for next io        

PRI out_ledport
'Handle writes to the LED port
'No, SIMH never had this. I added it for the Propeller demo board.
  'outa[16..23] := io_data                         'Output the port data  (if re-enabled do #ifdef)
  TV.str(string("LED port = "))
  TV.hex(io_data,2)
  TV.out(13)

PRI out_simhport
'Handle writes to the SIMH simulator control port


PRI out_condata
'Handle writes to the console data output port
  'data := data & $7F                             'Use this for 4KBASIC which spits out crap
  if io_data <> 10
    TV.out(io_data)


PRI out_constat
'Handle writes to the console command port
  if io_data == creset
    'TV.str(string("Console reset"))
    'TV.out(13)
  else
    'TV.str(string("Unknown console status write $"))
    'TV.hex(data,2)
    'TV.out(13)


PRI out_punstat
'Handle writes to the punch device command port
  if io_data == creset
    'TV.str(string("Punch reset"))
    'TV.out(13)
  else
    'TV.str(string("Unknown punch status write $"))
    'TV.hex(data,2)
    'TV.out(13)             


PRI out_selout
'Handle writes to the floppy disk drive select port
  if io_data & %10000000
    altair_drive := -1
  else
    altair_drive := io_data & %00001111


PRI out_dskcon
'Handle writes to the floppy disk control port
  if (io_data & cstepin) > 0                     'Step head in command ?
    if altair_track[altair_drive] < tracks
      altair_track[altair_drive]++
  if (io_data & cstepot) > 0                     'Step head out command ?
    if altair_track[altair_drive] > 0
      altair_track[altair_drive]--
          
  if (io_data & cload) > 0                       'Load head to drive surface command ?

  if (io_data & cuload) > 0                      'Unload head from drive surface command ?

  if (io_data & cwrseq) > 0                      ' 'start write enable sequence' command
    altair_bidx := 0
    altair_write_enable := 1


PRI out_dskwrit | r
'Handle writes to the floppy disk write data port
  if altair_bidx < asecsiz
    disk_buff[altair_bidx++] := io_data
    if altair_bidx == asecsiz
      'Write the buffer to the SD card or RamDisk

#ifdef TriBladeProp

#ifdef RamDisk
      sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
      err := tbp2.DoCmd("W", @disk_buff[3], sd_block*128 + disk_offset, 128)   'works RR20090404
#else
'----------------------------------------------------------------------------------------------
'The following code resolves conflicts by stopping and restarting the drivers !!!!! needs fixing
'----------------------------------------------------------------------------------------------
      r := tbp2.stop                                    '<=== stop the tbp ram driver   !!!!
      CheckError(r)
       
      r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)      'init the SD
      CheckError(r)
       
      ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
{
      sd_block := (altair_track[altair_drive] * 32) + altair_sector
      if altair_drive == 0
        sd_block := sd_block + drive0base
      else
        sd_block := sd_block + drive1base
}
      sd_block := (drive_base[altair_drive]) + (altair_track[altair_drive] * 32) + altair_sector
      err := sd.writeSDCard(sd_block, @disk_buff[3], 128) 'write a sector
'     CheckError(r)
       
      r := sd.stopSDcard                                'stop the SD (free up pins, but keep cog running)
      CheckError(r)
       
      r := tbp2.start                                     '<=== restart the tbp ram driver
      CheckError(r)
  '----------------------------------------------------------------------------------------------
#endif

#else
      sd_block := (altair_drive * tracks * 32) + (altair_track[altair_drive] * 32) + altair_sector
      err := \disk.writeblock(sd_block, @disk_buff)
#endif

      if err
        TV.str(string("Disk Write "))
        DisplayError(err)
        DisplayDiskDTS

      altair_write_enable := 0


PRI DisplayDiskDTS
      TV.str(string("drive = "))
      TV.hex(altair_drive, 2)
      TV.str(string(" track = "))
      TV.hex(altair_track, 2)
      TV.str(string(" sector = "))
      TV.hex(altair_sector, 2)
      TV.out(13)
          

PRI print_regs
'Print the Z80 registers
  TV.str(string("PC="))
  TV.hex(pc_reg, 4)
  TV.out(" ")

  TV.hex(byte[ram_base + pc_reg], 2)
  TV.out(" ")

  TV.str(string("AF="))
  TV.hex(af_reg, 4)
  TV.out(" ")
      
  TV.str(string("BC="))
  TV.hex(bc_reg, 4)
  TV.out(" ")
      
  TV.str(string("DE="))
  TV.hex(de_reg, 4)
  TV.out(" ")

  TV.str(string("HL="))
  TV.hex(hl_reg, 4)
  TV.out(" ")
      
  TV.str(string("IX="))
  TV.hex(ix_reg, 4)
  TV.out(" ")

  TV.str(string("IY="))
  TV.hex(iy_reg, 4)
  TV.out(" ")

  TV.str(string("SP="))
  TV.hex(sp_reg, 4)
  TV.out(13)

#ifdef Display_Sectors
PRI DisplaySector(s) | ix
'Read and display a 128B sector (SRAM, no offset)
  err := tbp2.DoCmd("R", @disk_buff, s*128, 128)
  TV.str(string("Sector="))
  TV.hex(s,8)
  TV.str(string(" "))
  DisplayError(err)
  repeat ix from 0 to 127
    TV.hex(disk_buff[ix],2)
    TV.str(string(" "))
  TV.str(string(13))
#endif

  